from typing import Final

from .core import Formula, evaluate
from .expressions import Predicate
from .trace import Trace
from ._banquo_impl import Top as _Top
from ._banquo_impl import Bottom as _Bottom

Top: Final[_Top] = _Top()
Bottom: Final[_Bottom] = _Bottom()

__all__ = [
    "Bottom",
    "Formula",
    "Predicate",
    "Top",
    "Trace",
    "evaluate",
]



from __future__ import annotations

from typing import Final

import pytest
import typing_extensions

import banquo
import banquo.operators as ops

pytestmark = pytest.mark.conformance

EPSILON: Final[float] = 1.0e-5
State: typing_extensions.TypeAlias = dict[str, float]
Trace: typing_extensions.TypeAlias = banquo.Trace[State]
Predicate: typing_extensions.TypeAlias = banquo.Predicate


@pytest.fixture
def trace() -> Trace:
    entries = [
        (0.0000, 0.0000),
        (0.3947, 0.5881),
        (0.7587, 1.1068),
        (1.0660, 1.4967),
        (1.2998, 1.7169),
        (1.4546, 1.7508),
        (1.5377, 1.6075),
        (1.5675, 1.3204),
        (1.5708, 0.9412),
        (1.5787, 0.5313),
        (1.6216, 0.1525),
        (1.7242, -0.1431),
        (1.9019, -0.3207),
        (2.1583, -0.368),
        (2.4844, -0.2963),
        (2.8603, -0.1383),
        (3.2583, 0.0582),
        (3.6471, 0.2386),
        (3.9968, 0.3511),
        (4.2840, 0.3561),
        (4.4947, 0.2326),
        (4.6273, -0.017),
        (4.6925, -0.3667),
        (4.7114, -0.7708),
        (4.7128, -1.1705),
        (4.7280, -1.5029),
        (4.7861, -1.7113),
        (4.9095, -1.7537),
        (5.1104, -1.6104),
        (5.3886, -1.2874),
        (5.7317, -0.816),
    ]

    return banquo.Trace({time: {"x": value} for time, value in entries})


@pytest.fixture
def p1() -> Predicate:
    return banquo.Predicate({"x": -1.0}, 2.0)


@pytest.fixture
def p2() -> Predicate:
    return banquo.Predicate({"x": 1.0}, 2.0)


def test_case01(trace: Trace, p1: Predicate, p2: Predicate):
    formula = ops.Or(p1, p2)
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(2.0, abs=EPSILON)


def test_case02(trace: Trace, p1: Predicate):
    formula = ops.Eventually(p1)
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(3.7508, abs=EPSILON)


def test_case03(trace: Trace, p2: Predicate):
    formula = ops.Always(p2)
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(0.2492, abs=EPSILON)


def test_case04(trace: Trace, p1: Predicate):
    formula = ops.Next(p1)
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(2.5881, abs=EPSILON)


def test_case05(trace: Trace, p1: Predicate):
    formula = ops.Next(ops.Next(p1))
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(3.1068, abs=EPSILON)


def test_case06(trace: Trace, p1: Predicate):
    formula = ops.Next(ops.Next(ops.Next(p1)))
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(3.4967, abs=EPSILON)


def test_case07(trace: Trace, p1: Predicate, p2: Predicate):
    formula = ops.Always(ops.Eventually(ops.And(p2, ops.Eventually(p1))))
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(1.184, abs=EPSILON)


def test_case08(trace: Trace, p1: Predicate, p2: Predicate):
    formula = ops.Implies(p1, p2)
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(2.0, abs=EPSILON)


def test_case09(trace: Trace, p1: Predicate, p2: Predicate):
    formula = ops.And(ops.Eventually(p1), ops.Always(ops.Implies(p1, ops.Eventually(p2))))
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(2.816, abs=EPSILON)


def test_case10(trace: Trace, p1: Predicate):
    formula = ops.Always(ops.Implies(p1, ops.Eventually(ops.Not(p1))))
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(-1.184, abs=EPSILON)


def test_case11(trace: Trace, p1: Predicate):
    formula = ops.Always(ops.Or(ops.Not(p1), ops.Eventually(ops.Always(ops.Not(p1)))))
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(-1.184, abs=EPSILON)


def test_case12(trace: Trace, p1: Predicate):
    formula = ops.Always(ops.Next(p1))
    result = banquo.evaluate(formula, trace)

    assert result == banquo.Bottom


def test_case13(trace: Trace, p1: Predicate):
    formula = ops.Eventually(ops.Next(p1))
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(3.7508, abs=EPSILON)


def test_case14(trace: Trace, p1: Predicate):
    formula = ops.Always(ops.Next(ops.Next(p1)))
    result = banquo.evaluate(formula, trace)

    assert result == banquo.Bottom


def test_case15(trace: Trace, p1: Predicate):
    formula = ops.Next(ops.Eventually(ops.Next(p1)))
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(3.7508, abs=EPSILON)


"""
fn case16() -> TestResult {
    let phi = Until::new(p1(), p2());
    let trace = get_trace();
    let rho = evaluate(&trace, phi)?;

    assert_relative_eq!(rho, 2.0, epsilon = EPSILON);

    Ok(())
}

#[test]
fn case17() -> TestResult {
    let phi = Until::new(p1(), Until::new(p2(), p1()));
    let trace = get_trace();
    let rho = evaluate(&trace, phi)?;

    assert_relative_eq!(rho, 2.0, epsilon = EPSILON);

    Ok(())
}
"""


def test_case23(trace: Trace, p1: Predicate, p2: Predicate):
    formula = ops.Not(ops.Eventually.with_bounds((0.0, 3.5), ops.Not(ops.And(p2, p1))))
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(0.2492, abs=EPSILON)


def test_case24(trace: Trace, p1: Predicate):
    formula = ops.Not(ops.Eventually.with_bounds((0.0, 1.0), ops.Not(p1)))
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(2.0, abs=EPSILON)


def test_case25(trace: Trace, p1: Predicate):
    formula = ops.Eventually.with_bounds((0.1, 30.0), p1)
    result = banquo.evaluate(formula, trace)

    assert isinstance(result, float)
    assert banquo.evaluate(formula, trace) == pytest.approx(3.7508, abs=EPSILON)



from __future__ import annotations

from collections.abc import Mapping

from ._banquo_impl import Predicate as _Predicate
from .operators import OperatorMixin
from .core import EnsureOutput


class Predicate(EnsureOutput[dict[str, float], float], OperatorMixin):
    """A temporal logic expression of the form ax <= b.

    In the interest of clarity, and to reduce the overhead of needing to ensure the variables
    in each state of the trace are in the correct order, we use dictionaries to represent
    both the system state and the *a* vector. Checking for membership is done on the coefficient
    side, which means that the state may contain an arbitrary number of additional variables so
    long as it contains all of the variables for which a coefficient has been defined.

    Args:
        coefficients: The a vector provided as a dictionary of variables and their coefficients
        constant: The b value on the right side of the inequality
    """

    def __init__(self, coefficients: Mapping[str, float], constant: float):
        super().__init__(_Predicate(dict(coefficients), constant))



from __future__ import annotations

from typing import Protocol, TypeVar

from typing_extensions import Self, override

from ._banquo_impl import Trace as _Trace
from .trace import Trace

S = TypeVar("S", contravariant=True)
M = TypeVar("M", covariant=True)


class Formula(Protocol[S, M]):
    """A Formula evaluates a `Trace` of system states into a `Trace` of metric values.

    This protocol accepts two type parameters, S and M. S represents the type of the
    state values this implementation can evaluate, and M represents the type of the
    metric this implementation produces.

    In general, there are two kinds of implementations for this interface. The first
    we refer to as *expressions*, which evaluate the trace directly into metric values.
    The second are *operators*, which generally delegate evaluation to one or more
    sub-formulas and then combine the resultant traces from each subformula into a single
    trace.
    """

    def evaluate(self, trace: Trace[S]) -> _Trace[M]:
        """Evaluate the system states into metric values.

        The trace returned by this method should contain the same times as the input trace.

        Args:
            trace: The trace of system states

        Returns:
            A trace of metric values computed from each system state
        """
        ...


class SupportsNeg(Protocol):
    def __neg__(self) -> Self: ...


class SupportsLE(Protocol):
    def __le__(self, value: Self, /) -> bool: ...


class SupportsGE(Protocol):
    def __ge__(self, value: Self, /) -> bool: ...


class SupportsNegGE(SupportsNeg, SupportsGE, Protocol): ...


class EnsureInput(Formula[S, M]):
    """Wrapper to convert the input trace to a `Trace` instance if it is not already."""

    def __init__(self, inner: Formula[S, M]):
        self.inner: Formula[S, M] = inner

    @override
    def evaluate(self, trace: _Trace[S]) -> _Trace[M]:
        return self.inner.evaluate(trace if isinstance(trace, Trace) else Trace(trace))


class EnsureOutput(Formula[S, M]):
    """Wrapper to convert traces returned from rust-implemented operators into Trace values."""

    def __init__(self, inner: Formula[S, M]):
        self.inner: Formula[S, M] = inner

    @override
    def evaluate(self, trace: Trace[S]) -> Trace[M]:
        result: _Trace[M] = self.inner.evaluate(trace)
        return result if isinstance(result, Trace) else Trace(result)


def evaluate(formula: Formula[S, M], trace: Trace[S]) -> M:
    """Evaluate a trace into a metric.

    In general, when we evaluate formulas we only care about the metric value at the earliest time
    in the trace.

    Args:
        formula: The formula to use for evaluation
        trace: The trace of system states

    Returns:
        The first value in the metric trace returned by the formula.

    Raises:
        ValueError: If the provided formula evaluates the given input trace to an empty metric trace.
    """

    result = formula.evaluate(trace)

    try:
        return next(iter(result.states()))
    except StopIteration:
        raise ValueError("Provided formula evaluated to an empty trace.")



from collections.abc import Iterable
from typing import Any, Generic, TypeVar

from typing_extensions import Self, override

from .core import Formula
from .operators import Bounds, M, M_neg, M_le, M_ge, M_neg_ge

class PanicException(Exception): ...

T = TypeVar("T", covariant=True)

class Top(Any):  # pyright: ignore[reportAny, reportExplicitAny]
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    @override
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...

class Bottom(Any):  # pyright: ignore[reportAny, reportExplicitAny]
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    @override
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...

class Trace(Generic[T]):
    def __new__(cls, elements: dict[float, T] | Trace[T]) -> Self: ...
    def __getitem__(self, value: float) -> T: ...
    def times(self) -> Iterable[float]: ...
    def states(self) -> Iterable[T]: ...
    def at_time(self, time: float) -> T | None: ...

class Predicate(Formula[dict[str, float], float]):
    def __new__(cls, coefficients: dict[str, float], constant: float) -> Self: ...
    @override
    def evaluate(self, trace: Trace[dict[str, float]]) -> Trace[float]: ...

S = TypeVar("S", contravariant=True)

class Not(Formula[S, M_neg]):
    def __new__(cls, subformula: Formula[S, M_neg]) -> Self: ...
    @override
    def evaluate(self, trace: Trace[S]) -> Trace[M_neg]: ...

class And(Formula[S, M_le]):
    def __new__(cls, lhs: Formula[S, M_le], rhs: Formula[S, M_le]) -> Self: ...
    @override
    def evaluate(self, trace: Trace[S]) -> Trace[M_le]: ...

class Or(Formula[S, M_ge]):
    def __new__(cls, lhs: Formula[S, M_ge], rhs: Formula[S, M_ge]) -> Self: ...
    @override
    def evaluate(self, trace: Trace[S]) -> Trace[M_ge]: ...

class Implies(Formula[S, M_neg_ge]):
    def __new__(cls, lhs: Formula[S, M_neg_ge], rhs: Formula[S, M_neg_ge]) -> Self: ...
    @override
    def evaluate(self, trace: Trace[S]) -> Trace[M_neg_ge]: ...

class Next(Formula[S, M]):
    def __new__(cls, subformula: Formula[S, M]) -> Self: ...
    @override
    def evaluate(self, trace: Trace[S]) -> Trace[M]: ...

class Always(Formula[S, M_le]):
    def __new__(cls, bounds: Bounds | None, subformula: Formula[S, M_le]) -> Self: ...
    @override
    def evaluate(self, trace: Trace[S]) -> Trace[M_le]: ...

class Eventually(Formula[S, M_ge]):
    def __new__(cls, bounds: Bounds | None, subformula: Formula[S, M_ge]) -> Self: ...
    @override
    def evaluate(self, trace: Trace[S]) -> Trace[M_ge]: ...



from __future__ import annotations

from collections.abc import Iterable, Iterator, Mapping
from typing import TypeVar

from typing_extensions import override

from ._banquo_impl import Trace as _Trace

T = TypeVar("T", covariant=True)
U = TypeVar("U", covariant=True)


def _iter_eq(lhs: Iterable[object], rhs: Iterable[object]) -> bool:
    return list(lhs) == list(rhs)


class MismatchedTimesStates(Exception):
    def __init__(self):
        super().__init__("Times and states iterables must contain equal number of elements")


class Trace(_Trace[T], Iterable[tuple[float, T]]):
    """A set of values and their associated times.

    Iterating over a trace yields (time, value) pairs.

    Args:
        elements: The set of times and their associated values
    """

    def __new__(cls, elements: Mapping[float, T] | _Trace[T]):
        return super().__new__(cls, elements if isinstance(elements, _Trace) else dict(elements))

    @override
    def __str__(self) -> str:
        return str({time: state for time, state in self})

    @override
    def __iter__(self) -> Iterator[tuple[float, T]]:
        return zip(super().times(), super().states())

    @override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _Trace):
            return NotImplemented

        return _iter_eq(self.times(), other.times()) and _iter_eq(self.states(), other.states())

    @override
    def times(self) -> Iterable[float]:
        """Iterator over the times of the trace."""

        return iter(super().times())

    @override
    def states(self) -> Iterable[T]:
        """Iterator over the states of the trace."""

        return iter(super().states())

    @staticmethod
    def from_timed_states(times: Iterable[float], states: Iterable[U]) -> Trace[U]:
        """Construct a trace from a set of times and a set of states.

        Args:
            times: The times of the trace
            states: The state for each time

        Returns:
            A `Trace` instance from the times and states

        Raises:
            MismatchedTimesStates: If the number of time values does not match the number of state values.
        """

        times_ = list(times)
        states_ = list(states)

        if len(times_) != len(states_):
            raise MismatchedTimesStates()

        return Trace({time: state for time, state in zip(times_, states_)})



from __future__ import annotations

import typing

from typing_extensions import TypeAlias, override

from ._banquo_impl import And as _And
from ._banquo_impl import Always as _Always
from ._banquo_impl import Eventually as _Eventually
from ._banquo_impl import Implies as _Implies
from ._banquo_impl import Next as _Next
from ._banquo_impl import Not as _Not
from ._banquo_impl import Or as _Or
from ._banquo_impl import PanicException
from .core import (
    Formula,
    EnsureInput,
    SupportsNeg,
    SupportsLE,
    SupportsGE,
    EnsureOutput,
    SupportsNegGE,
)
from .trace import Trace

Bounds: TypeAlias = tuple[float, float]

S = typing.TypeVar("S")
M = typing.TypeVar("M", covariant=True)
M_neg = typing.TypeVar("M_neg", bound=SupportsNeg, covariant=True)
M_le = typing.TypeVar("M_le", bound=SupportsLE, covariant=True)
M_ge = typing.TypeVar("M_ge", bound=SupportsGE, covariant=True)
M_neg_ge = typing.TypeVar("M_neg_ge", bound=SupportsNegGE, covariant=True)


class OperatorMixin:
    def and_(self: Formula[S, M_le], other: Formula[S, M_le]) -> And[S, M_le]:
        return And(self, other)


class MetricAttributeError(AttributeError):
    def __init__(self, missing_method: str):
        super().__init__(f"Metric must implement {missing_method} method")


class Operator(EnsureOutput[S, M], OperatorMixin):
    def __init__(self, subformula: Formula[S, M], required_method: str):
        super().__init__(subformula)
        self.required_method: str = required_method

    @override
    def evaluate(self, trace: Trace[S]) -> Trace[M]:
        try:
            return super().evaluate(trace)
        except PanicException as e:
            raise MetricAttributeError(self.required_method) from e


def _inner_or_wrap(formula: Formula[S, M]) -> Formula[S, M]:
    if isinstance(formula, EnsureOutput):
        return formula.inner

    return EnsureInput(formula)


class Not(Operator[S, M_neg]):
    """The first-order negation operator.

    This operator holds only if the subformula does not hold. This operator requires the metric
    value returned by its subformula to implement the `__neg__` method.

    Args:
        subformula: The formula to negate
    """

    def __init__(self, subformula: Formula[S, M_neg]):
        super().__init__(_Not(_inner_or_wrap(subformula)), "__neg__")


class And(Operator[S, M_le]):
    """The first-order conjunction operator.

    This operator holds if both the left and right subformulas hold. This operator requires
    the metric value returned by its subformulas to implement the `__le__` method.

    Args:
        lhs: The left side subformula
        rhs: The right side subformula
    """

    def __init__(self, lhs: Formula[S, M_le], rhs: Formula[S, M_le]):
        super().__init__(_And(_inner_or_wrap(lhs), _inner_or_wrap(rhs)), "__le__")


class Or(Operator[S, M_ge]):
    """The first-order disjunction operator.

    This operator holds if either the left or right subformula holds. This operator requires the
    metric value returned by its subformulas to implement the `__ge__` method.

    Args:
        lhs: The left side subformula
        rhs: The right side subformula
    """

    def __init__(self, lhs: Formula[S, M_ge], rhs: Formula[S, M_ge]):
        super().__init__(_Or(_inner_or_wrap(lhs), _inner_or_wrap(rhs)), "__ge__")


class Implies(Operator[S, M_neg_ge]):
    """The first-order disjunction operator.

    This operator holds if the left and right subformula holds, or if the left subformula does not
    hold. This operator requires the metric value returned by its subformulas to implement the
    `__neg__` and `__ge__` methods.

    Args:
        lhs: The left side subformula
        rhs: The right side subformula
    """

    def __init__(self, lhs: Formula[S, M_neg_ge], rhs: Formula[S, M_neg_ge]):
        super().__init__(_Implies(_inner_or_wrap(lhs), _inner_or_wrap(rhs)), "__neg__ and __ge__")


class Next(Operator[S, M]):
    """The temporal shift operator.

    This operator holds if the subformula holds at the next time in the future.

    Args:
        subformula: The subformula to shift
    """

    def __init__(self, subformula: Formula[S, M]):
        super().__init__(_Next(subformula), "")


S_ = typing.TypeVar("S_")
M_le_ = typing.TypeVar("M_le_", bound=SupportsLE, covariant=True)


class Always(Operator[S, M_le]):
    """The temporal Globally operator.

    This operator holds if its subformula holds for all times in the future. This operator
    requires that the metric returned from the subformula to implement the `__le__` method.

    Args:
        subformula: The subformula to evaluate
    """

    def __init__(self, subformula: Formula[S, M_le]):
        if isinstance(subformula, _Always):
            inner = subformula
        else:
            inner = _Always(None, _inner_or_wrap(subformula))

        super().__init__(inner, "__le__")

    @staticmethod
    def with_bounds(bounds: Bounds, subformula: Formula[S_, M_le_]) -> Always[S_, M_le_]:
        """Construct a bounded Always operator.

        A bounded operator sets a time horizon for evaluation of the subformula. In this case, for a
        bounded interval (lo, hi), the operator will only consider future times in the range
        (t+lo, t+hi) at every time t.

        Args:
            bounds: The time interval to restrict evaluation to
            subformula: The subformula to evaluate

        Returns:
            An Always operator with a restricted evaluation interval.
        """

        return Always(_Always(bounds, _inner_or_wrap(subformula)))


M_ge_ = typing.TypeVar("M_ge_", bound=SupportsGE, covariant=True)


class Eventually(Operator[S, M_ge]):
    """The temporal Finally operator.

    This operator holds if its subformula holds for at any time in the future. This operator
    requires that the metric returned from the subformula to implement the `__ge__` method.

    Args:
        subformula: The subformula to evaluate
    """

    def __init__(self, subformula: Formula[S, M_ge]):
        if isinstance(subformula, _Eventually):
            inner = subformula
        else:
            inner = _Eventually(None, _inner_or_wrap(subformula))

        super().__init__(inner, "__le__")

    @staticmethod
    def with_bounds(bounds: Bounds, subformula: Formula[S_, M_ge_]) -> Eventually[S_, M_ge_]:
        """Construct a bounded Eventually operator.

        A bounded operator sets a time horizon for evaluation of the subformula. In this case, for a
        bounded interval (lo, hi), the operator will only consider future times in the range
        (t+lo, t+hi) at every time t.

        Args:
            bounds: The time interval to restrict evaluation to
            subformula: The subformula to evaluate

        Returns:
            An Eventually operator with a restricted evaluation interval.
        """

        return Eventually(_Eventually(bounds, _inner_or_wrap(subformula)))



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar

import pytest
import typing_extensions

from banquo import Bottom, Trace, operators
from banquo.core import Formula

pytestmark = pytest.mark.unit
T = TypeVar("T")


@dataclass()
class BadMetric:
    value: float


@dataclass()
class GoodMetric(BadMetric):
    @typing_extensions.override
    def __eq__(self, other: object) -> bool:
        if isinstance(other, GoodMetric):
            return self.value == other.value

        if isinstance(other, float):
            return self.value == other

        return NotImplemented

    def __neg__(self) -> GoodMetric:
        return GoodMetric(-self.value)

    def __le__(self, other: object, /) -> bool:
        if not isinstance(other, GoodMetric):
            return NotImplemented

        return self.value <= other.value

    def __ge__(self, other: object, /) -> bool:
        if not isinstance(other, GoodMetric):
            return NotImplemented

        return self.value >= other.value


class Const(Formula[T, T]):
    @typing_extensions.override
    def evaluate(self, trace: Trace[T]) -> Trace[T]:
        return trace


class UnaryTest:
    @pytest.fixture
    def input(self) -> Trace[float]:
        return Trace({
            0.0: 1.0,
            1.0: 1.1,
            2.0: 1.2,
            3.0: 1.3,
            4.0: 1.2,
            5.0: 1.5,
        })

    @pytest.fixture
    def good_trace(self, input: Trace[float]) -> Trace[GoodMetric]:
        return Trace({time: GoodMetric(state) for time, state in input})

    @pytest.fixture
    def bad_trace(self, input: Trace[float]) -> Trace[BadMetric]:
        return Trace({time: BadMetric(state) for time, state in input})


class TestNegation(UnaryTest):
    @pytest.fixture
    def expected(self, input: Trace[float]) -> Trace[float]:
        return Trace({time: -state for time, state in input})

    def test_evaluation(self, input: Trace[float], expected: Trace[float]):
        formula = operators.Not(Const[float]())
        result = formula.evaluate(input)

        assert isinstance(result, Trace)
        assert result == expected

    def test_supported_metric(self, good_trace: Trace[GoodMetric], expected: Trace[float]):
        formula = operators.Not(Const[GoodMetric]())
        assert formula.evaluate(good_trace) == expected

    def test_unsupported_metric(self, bad_trace: Trace[BadMetric]):
        formula = operators.Not(Const[BadMetric]())  # pyright: ignore[reportArgumentType, reportUnknownVariableType]

        with pytest.raises(operators.MetricAttributeError):
            _ = formula.evaluate(bad_trace)  # pyright: ignore[reportUnknownVariableType]


L = TypeVar("L")
R = TypeVar("R")


class Left(Formula[tuple[L, R], L]):
    @typing_extensions.override
    def evaluate(self, trace: Trace[tuple[L, R]]) -> Trace[L]:
        assert isinstance(trace, Trace)
        return Trace({time: state[0] for time, state in trace})


class Right(Formula[tuple[L, R], R]):
    @typing_extensions.override
    def evaluate(self, trace: Trace[tuple[L, R]]) -> Trace[R]:
        return Trace({time: state[1] for time, state in trace})


class BinaryTest:
    @pytest.fixture
    def input(self) -> Trace[tuple[float, float]]:
        return Trace({
            0.0: (0.0, 1.0),
            1.0: (1.0, 0.0),
            2.0: (2.0, 4.0),
            3.0: (3.0, 6.0),
        })

    @pytest.fixture
    def good_trace(self, input: Trace[tuple[float, float]]) -> Trace[tuple[GoodMetric, GoodMetric]]:
        return Trace({time: (GoodMetric(state[0]), GoodMetric(state[1])) for time, state in input})

    @pytest.fixture
    def bad_trace(sefl, input: Trace[tuple[float, float]]) -> Trace[tuple[BadMetric, BadMetric]]:
        return Trace({time: (BadMetric(state[0]), BadMetric(state[1])) for time, state in input})


class TestConjunction(BinaryTest):
    @pytest.fixture
    def expected(self, input: Trace[tuple[float, float]]) -> Trace[float]:
        return Trace({time: min(state) for time, state in input})

    def test_evaluation(self, input: Trace[tuple[float, float]], expected: Trace[float]):
        formula = operators.And(Left[float, float](), Right[float, float]())
        result = formula.evaluate(input)

        assert isinstance(result, Trace)
        assert result == expected

    def test_supported_metric(
        self, good_trace: Trace[tuple[GoodMetric, GoodMetric]], expected: Trace[float]
    ):
        formula = operators.And(Left[GoodMetric, GoodMetric](), Right[GoodMetric, GoodMetric]())
        assert formula.evaluate(good_trace) == expected

    def test_unsupported_metric(self, bad_trace: Trace[tuple[BadMetric, BadMetric]]):
        formula = operators.And(Left[BadMetric, BadMetric](), Right[BadMetric, BadMetric]())  # pyright: ignore[reportArgumentType, reportUnknownVariableType]

        with pytest.raises(operators.MetricAttributeError):
            _ = formula.evaluate(bad_trace)  # pyright: ignore[reportUnknownVariableType]


class TestDisjunction(BinaryTest):
    @pytest.fixture
    def expected(self, input: Trace[tuple[float, float]]) -> Trace[float]:
        return Trace({time: max(state) for time, state in input})

    def test_evaluation(self, input: Trace[tuple[float, float]], expected: Trace[float]):
        formula = operators.Or(Left[float, float](), Right[float, float]())
        result = formula.evaluate(input)

        assert isinstance(result, Trace)
        assert result == expected

    def test_supported_metric(
        self, good_trace: Trace[tuple[GoodMetric, GoodMetric]], expected: Trace[float]
    ):
        formula = operators.Or(Left[GoodMetric, GoodMetric](), Right[GoodMetric, GoodMetric]())
        assert formula.evaluate(good_trace) == expected

    def test_unsupported_metric(self, bad_trace: Trace[tuple[BadMetric, BadMetric]]):
        formula = operators.Or(Left[BadMetric, BadMetric](), Right[BadMetric, BadMetric]())  # pyright: ignore[reportArgumentType, reportUnknownVariableType]

        with pytest.raises(operators.MetricAttributeError):
            _ = formula.evaluate(bad_trace)  # pyright: ignore[reportUnknownVariableType]


class TestImplication(BinaryTest):
    @pytest.fixture
    def expected(self, input: Trace[tuple[float, float]]) -> Trace[float]:
        return Trace({time: max(-state[0], state[1]) for time, state in input})

    def test_evaluation(self, input: Trace[tuple[float, float]], expected: Trace[float]):
        formula = operators.Implies(Left[float, float](), Right[float, float]())
        result = formula.evaluate(input)

        assert isinstance(result, Trace)
        assert result == expected

    def test_supported_metric(
        self, good_trace: Trace[tuple[GoodMetric, GoodMetric]], expected: Trace[float]
    ):
        formula = operators.Implies(Left[GoodMetric, GoodMetric](), Right[GoodMetric, GoodMetric]())
        result = formula.evaluate(good_trace)
        assert result == expected

    def test_unsupported_metric(self, bad_trace: Trace[tuple[BadMetric, BadMetric]]):
        formula = operators.Implies(Left[BadMetric, BadMetric](), Right[BadMetric, BadMetric]())  # pyright: ignore[reportArgumentType, reportUnknownVariableType]

        with pytest.raises(operators.MetricAttributeError):
            _ = formula.evaluate(bad_trace)  # pyright: ignore[reportUnknownVariableType]


class TestNext(UnaryTest):
    @pytest.fixture
    def expected(self) -> Trace[float]:
        return Trace({
            0.0: 1.1,
            1.0: 1.2,
            2.0: 1.3,
            3.0: 1.2,
            4.0: 1.5,
            5.0: Bottom,
        })

    def test_evaluation(self, input: Trace[float], expected: Trace[float]):
        formula = operators.Next(Const[float]())
        result = formula.evaluate(input)

        assert isinstance(result, Trace)
        assert result == expected

    def test_userdefined_metric(self, good_trace: Trace[GoodMetric], expected: Trace[float]):
        formula = operators.Next(Const[GoodMetric]())
        assert formula.evaluate(good_trace) == expected


class TestGlobally(UnaryTest):
    @pytest.fixture
    def expected(self) -> Trace[float]:
        return Trace({
            0.0: 1.0,
            1.0: 1.1,
            2.0: 1.2,
            3.0: 1.2,
            4.0: 1.2,
            5.0: 1.5,
        })

    def test_bounded_evaluation(self, input: Trace[float], expected: Trace[float]):
        formula = operators.Always(Const[float]())
        result = formula.evaluate(input)

        assert isinstance(result, Trace)
        assert result == expected

    def test_unbounded_evaluation(self, input: Trace[float]):
        formula = operators.Always.with_bounds((0.0, 2.0), Const[float]())
        expected = Trace({
            0.0: 1.0,
            1.0: 1.1,
            2.0: 1.2,
            3.0: 1.2,
            4.0: 1.2,
            5.0: 1.5,
        })
        result = formula.evaluate(input)

        assert isinstance(result, Trace)
        assert result == expected

    def test_supported_metric(self, good_trace: Trace[GoodMetric], expected: Trace[float]):
        formula = operators.Always(Const[GoodMetric]())
        assert formula.evaluate(good_trace) == expected

    def test_unsupported_metric(self, bad_trace: Trace[BadMetric]):
        formula = operators.Always(Const[BadMetric]())  # pyright: ignore[reportArgumentType, reportUnknownVariableType]

        with pytest.raises(operators.MetricAttributeError):
            _ = formula.evaluate(bad_trace)  # pyright: ignore[reportUnknownVariableType]


class TestFinally(UnaryTest):
    @pytest.fixture
    def expected(self) -> Trace[float]:
        return Trace({
            0.0: 1.5,
            1.0: 1.5,
            2.0: 1.5,
            3.0: 1.5,
            4.0: 1.5,
            5.0: 1.5,
        })

    def test_bounded_evaluation(self, input: Trace[float], expected: Trace[float]):
        formula = operators.Eventually(Const[float]())
        result = formula.evaluate(input)

        assert isinstance(result, Trace)
        assert result == expected

    def test_unbounded_evaluation(self, input: Trace[float]):
        formula = operators.Eventually.with_bounds((0.0, 2.0), Const[float]())
        expected = Trace({
            0.0: 1.2,
            1.0: 1.3,
            2.0: 1.3,
            3.0: 1.5,
            4.0: 1.5,
            5.0: 1.5,
        })
        result = formula.evaluate(input)

        assert isinstance(result, Trace)
        assert result == expected

    def test_supported_metric(self, good_trace: Trace[GoodMetric], expected: Trace[float]):
        formula = operators.Eventually(Const[GoodMetric]())
        assert formula.evaluate(good_trace) == expected

    def test_unsupported_metric(self, bad_trace: Trace[BadMetric]):
        formula = operators.Eventually(Const[BadMetric]())  # pyright: ignore[reportArgumentType, reportUnknownVariableType]

        with pytest.raises(operators.MetricAttributeError):
            _ = formula.evaluate(bad_trace)  # pyright: ignore[reportUnknownVariableType]



from __future__ import annotations

import itertools

import pytest

from banquo import Trace
from banquo.trace import MismatchedTimesStates

pytestmark = pytest.mark.unit


def test_trace():
    elements = {
        0.0: {"x": 0.0, "y": 1.0},
        1.0: {"x": 7.2, "y": 1.1},
        2.0: {"x": 13.3, "y": 1.2},
        3.0: {"x": 21.4, "y": 1.3},
        4.0: {"x": 30.1, "y": 1.2},
        5.0: {"x": 38.2, "y": 1.5},
    }

    times = elements.keys()
    states = elements.values()

    assert Trace(elements) == Trace.from_timed_states(times, states)  # Ensure constructor parity
    assert Trace(Trace(elements)) == Trace(elements)  # Ensure conversion is lossless

    short_states = itertools.islice(elements.values(), 4)  # Only produce 4 states

    with pytest.raises(MismatchedTimesStates):
        _ = Trace.from_timed_states(elements.keys(), short_states)
